from .snapshot import SNAPSHOT
from .utils import current_version, working_version
from .staging import STAGING
from .revert import REVERT
from .housing import HOUSE
import os
import sys

directory = '.svcs'
house = ''
housing = HOUSE(directory)
current_directory = os.getcwd()

def init_svcs() -> None:
  global house
  if not svcs_initialized():
    os.makedirs(directory, exist_ok=True)
    house = 'main'
    housing.new_house(house)
    print(f'SVCS initialized\n Current house: {house}')
    house_file = os.path.join(directory, "house.txt")
    all_house_file = os.path.join(directory, "all_house.txt")
    with open(house_file, "w") as f:
      f.write("main")
    with open(all_house_file, "w") as f:
      f.write("main")
  else:
    print("Already initialized")

def svcs_initialized() -> bool:
  return os.path.exists(directory)

def help_command() -> str:
  all_commands = '''
  List of all available commands
  - init  #To initialize svcs
  - house # View current house
  - house show #To view all the present houses
  - house <house-name> # Switch back to a house
  - house new <feature-branch>  # Create a new feature house
  - current # To get current version
  - snaps # To get total number of versions
  - ready # Stage your changes
  - snapshot # Create a snapshot
  - revert <version> # Revert to a version if needed
  '''
  return all_commands

def main() -> None:
  global house
  command = sys.argv
  command_length = len(command)
  if command_length < 2:
    print(help_command())
    return

  if command[1] == 'init':
    init_svcs()
    return

  if not svcs_initialized():
    print("SVCS not initialized please run \n svcs init")
    return

  house = housing.current_house()
  housing_path = os.path.join(directory, house)
  if command[1] == 'house':
    if command_length < 3:
      print(f'You are at {house}')
      return

    if command[2] == 'new':
      if command_length > 3:
        statement = housing.new_house(command[3])
        print(statement)
      else:
        print("Please define house name")
    elif command[2] == 'show':
      print(f"You have houses at\n {housing.all_house()}")
    else:
      statement = housing.move_house(command[2])
      print(statement)
  elif command[1] == 'ready':
    staging = STAGING(current_directory, directory, house)
    staging.ready()
  elif command[1] == 'snapshot':
    commiting = SNAPSHOT(current_directory, directory, house)
    commiting.snapshot()
  elif command[1] == 'revert':
    if command_length > 2:
      REVERT(directory, house, command[2]).soft_revert()
    else:
      print("Please define the version")
  elif command[1] == 'current':
    print(f'You are at version {working_version(housing_path)}')
  elif command[1] == 'snaps':
    print(f'There are {current_version(housing_path)} snaps')
  else:
    print('Unknown command!')
    print(help_command())

if __name__ == "__main__":
  main()


import os, pickle, shutil
from .ignore import dir_ignore, files_ignore
from pathlib import Path
from .utils import data_json_load, hashing

class STAGING:
  def __init__(self, current_dir :str, storage :str, house :str) -> None:
    self.directory = current_dir
    self.house_path = os.path.join(storage, house)
    self.temp = os.path.join(self.house_path, 'ready')
    self.json_file = os.path.join(self.house_path, 'data.json')
    self.__created_directory = set()
    self.__present_file_path = set()
    self.hash_data: dict = data_json_load(self.json_file)
    self.existing_files_with_hashes = set(self.hash_data.keys())

  def copying_files(self):
    for root, dirs, files in os.walk(self.directory):
      dirs[:] = [d for d in dirs if d not in dir_ignore]
      for file in files:
        if file in files_ignore:
          continue
        file_path = os.path.join(root, file)
        self.__present_file_path.add(file_path)
        if self.check_hash_exists(file_path):
          continue
        parent = Path(self.directory)
        child = Path(root)
        relative = child.relative_to(parent)
        folder = os.path.join(self.temp, str(relative))
        if relative not in self.__created_directory:
          os.makedirs(folder, exist_ok=True)
          self.__created_directory.add(relative)
        shutil.copy2(file_path, folder)

  def check_hash_exists(self, file_path):
    if self.hash_data:
      if self.hash_data.get(file_path):
        hash_value = hashing(file_path)
        if self.hash_data[file_path]['updated_hash'] == hash_value:
          return True
    return False

  def files_to_be_deleted(self) -> None:
    file_path = os.path.join(self.temp, 'to_be_deleted')
    if os.path.exists(file_path):
      os.remove(file_path)
    self.existing_files_with_hashes.remove("current_version")
    self.existing_files_with_hashes.remove("all_versions")
    set_of_files_to_be_deleted: set = self.existing_files_with_hashes - self.__present_file_path
    if set_of_files_to_be_deleted:
      with open(file_path, 'wb') as data_file:
        pickle.dump(set_of_files_to_be_deleted, data_file)

  def ready(self) -> None:
    self.copying_files()
    self.files_to_be_deleted()


import os, hashlib, pickle, shutil, gzip
from .utils import update_version, data_json_dump, data_json_load

class SNAPSHOT:
  def __init__(self, current_directory :str, directory :str, current_house :str) -> None:
    self.working_directory = os.path.join(directory, current_house)
    self.ready_directory = os.path.join(self.working_directory, 'ready')
    self.current_directory = current_directory
    self.snapshot_directory = os.path.join(self.working_directory, 'snapshot')
    self.json_file = os.path.join(self.working_directory, 'data.json')
    self._current_ver: None = None
    self._int_current_ver: None = None
    self._hash_data: None = None

  @property
  def current_ver(self):
    if self._current_ver is None:
      self._current_ver = update_version(self.working_directory)
    return self._current_ver

  @property
  def int_current_ver(self):
    if self._int_current_ver is None:
      self._int_current_ver = int(self.current_ver)
    return self._int_current_ver

  @property
  def hash_data(self):
    if self._hash_data is None:
      self._hash_data = data_json_load(self.json_file)
    return self._hash_data

  def if_directory_empty(self) -> bool:
    if not os.path.isdir(self.ready_directory):
      print(f"Error: '{self.ready_directory}' is not a valid directory.")
      return False

    with os.scandir(self.ready_directory) as it:
      return next(it, None) is None

  def files_deleted(self, file_path: str)-> None:
    with open(file_path, 'rb') as f:
      deleted: set = pickle.load(f)
    for file in deleted:
      self.hash_data[file]["deleted_in"] = self.int_current_ver

  @staticmethod
  def file_snapshot(file_path) -> (str, bytes):
    snapshot_hash = hashlib.sha256()
    with open(file_path, 'rb') as f:
      snapshot_data = f.read()
    snapshot_hash.update(snapshot_data)
    hash_digest = snapshot_hash.hexdigest()
    return hash_digest, snapshot_data

  def update_data_file(self, file_path, hash_digest):
    if self.hash_data.__contains__(file_path):
      self.hash_data[file_path]["updated_hash"] = hash_digest
      self.hash_data[file_path]["all_hashes"][self.current_ver] = hash_digest
    else:
      data = dict({"added_in": 0, "deleted_in": 0, "updated_hash": "", "all_hashes": {}})
      data["added_in"] = int(self.current_ver)
      data["updated_hash"] = hash_digest
      data["all_hashes"][self.current_ver] = hash_digest
      self.hash_data[file_path] = data

  def create_snapshot(self) -> None:
    for root, dirs, files in os.walk(self.ready_directory):
      for file in files:
        file_path = os.path.join(root, file)
        if file == 'to_be_deleted':
          self.files_deleted(file_path)
          continue
        hash_digest, snapshot_data = self.file_snapshot(file_path)
        file_path = file_path.replace(self.ready_directory, self.current_directory)
        self.update_data_file(file_path, hash_digest)
        save_file = os.path.join(self.snapshot_directory, hash_digest)
        with gzip.open(save_file, 'wb') as snap_file:
          pickle.dump(snapshot_data, snap_file)

  def empty_ready_folder(self) -> None:
    shutil.rmtree(self.ready_directory)
    os.makedirs(self.ready_directory)

  def snapshot(self) -> None:
    if not self.if_directory_empty():
      _ = self.current_ver
      _ = self.hash_data
      self.create_snapshot()
      data_json_dump(self.json_file, self.hash_data)
      print(f'Snapshot created')
      self.empty_ready_folder()
    else:
      print("Nothing to Snapshot")


import os
from .utils import data_json_load, data_json_dump, add_files, delete_files

class REVERT:
  def __init__(self, directory :str, house :str, version :str) -> None:
    self.housing_path = os.path.join(directory, house)
    self.root_path = os.path.join(self.housing_path, 'snapshot')
    self.json_path = os.path.join(self.housing_path, 'data.json')
    self.json_data: dict = data_json_load(self.json_path)
    self.version = version
    self.version_int = int(self.version)

  def revert(self) -> None:
    for file in self.json_data:
      if file == 'current_version':
        self.json_data['current_version'] = self.version_int
        continue
      elif file == 'all_versions':
        continue
      elif ((self.json_data[file]["added_in"] > self.version_int)
            or (self.json_data[file]["deleted_in"] < self.version_int)
            and (self.json_data[file]["deleted_in"] != 0)):
        delete_files(file)
      else:
        if self.json_data[file]["all_hashes"].__contains__(self.version):
          file_hash = self.json_data[file]["all_hashes"][self.version]
        else:
          file_hash = self.get_older_file_hash(file)
        add_files(file, self.root_path, file_hash)
        self.json_data[file]["updated_hash"] = file_hash
    data_json_dump(self.json_path, self.json_data)

  def get_older_file_hash(self, file) -> str:
    file_hash = ""
    list_data: dict = self.json_data[file]["all_hashes"]
    available_versions = list(map(int, list(list_data.keys())))
    for available in reversed(available_versions):
      if available < self.version_int:
        file_hash = list_data[f'{available}']
        break
    return file_hash

  def revert_checks(self) -> bool:
    if int(self.version) not in self.json_data['all_versions']:
      print('Snapshot does not exist.')
      return False
    if int(self.version) == self.json_data['current_version']:
      print(f"Already at {self.version}")
      return False
    return True

  def soft_revert(self) -> None:
    if self.revert_checks():
      self.revert()

  def hard_revert(self) -> None:
    self.revert()


# Cross-platform ignore patterns
dir_ignore = [
    ".venv", ".git", ".idea", "__pycache__", "venv",
    "Version-control", ".svcs", "svcs",
    # Windows specific
    "node_modules", ".vs", "bin", "obj",
    # macOS specific  
    ".DS_Store.d",
    # Linux specific
    ".cache"
]

files_ignore = [
    # macOS specific
    ".DS_Store", 
    # Windows specific
    "Thumbs.db", "desktop.ini",
    # Linux specific
    ".directory",
    # Cross-platform
    ".gitignore", ".gitkeep"
]


import os
from .revert import REVERT
from .utils import working_version, data_json_dump

class HOUSE:
  def __init__(self, svcs_directory :str) -> None:
    self.house_name :str = ''
    self.working_location: str = ''
    self.svcs_directory :str = svcs_directory
    self.house_file :str = os.path.join(svcs_directory, "house.txt")
    self.all_house_file :str = os.path.join(svcs_directory, "all_house.txt")

  def new_house(self, house_name) -> str:
    self.house_name: str = house_name
    self.working_location = os.path.join(self.svcs_directory, house_name)
    if os.path.exists(self.working_location):
      return f'{house_name} already exists please choose other name'
    else:
      os.makedirs(self.working_location)
      self.generate_rooms()
      self.update_house(house_name)
      return f'You are at {house_name}'

  def generate_rooms(self) -> None:
    path = self.working_location
    snapshot = os.path.join(path, 'snapshot')
    ready = os.path.join(path, 'ready')
    data_file_path = os.path.join(path, "data.json")
    os.makedirs(snapshot)
    os.makedirs(ready)
    content = {"current_version": 0, "all_versions": []}
    data_json_dump(data_file_path, content)

  def current_house(self) -> str:
    try:
      house_file = os.path.join(self.svcs_directory, 'house.txt')
      house = open(house_file, 'r')
      current = house.read()
      house.close()
      return current
    except FileNotFoundError:
      print('File not found')
      return 'Point of no return'

  def update_house(self, house_name: str) -> None:
    self.update_current_house(house_name)
    with open(self.all_house_file, "a") as f:
      f.write(f"\n{house_name}")

  def update_current_house(self, house_name):
    self.house_name: str = house_name
    with open(self.house_file, "w") as f:
      f.write(house_name)

  def move_house(self, house_name: str) -> str:
    self.working_location = os.path.join(self.svcs_directory, house_name)
    if os.path.exists(self.working_location):
      self.update_current_house(house_name)
      version = working_version(self.working_location)
      REVERT(self.svcs_directory, self.house_name, version).hard_revert()
      return f'You are at {self.house_name}'
    else:
      return 'House does not exists'

  def all_house(self) -> str:
    house = open(self.all_house_file, 'r')
    all_houses = house.read().strip()
    house.close()
    return all_houses



import os, json, hashlib, pickle, gzip

def current_version(directory :str) -> str:
  version_file_path = os.path.join(directory, 'data.json')
  with open(version_file_path, 'r') as version_file:
    file_content = json.load(version_file)
    versions = file_content["all_versions"]
    if len(versions):
      version = file_content["all_versions"][-1]
    else:
      version = 0
  return version

def update_version(directory :str) -> str:
  version = current_version(directory)
  version = int(version) + 1
  version_file_path = os.path.join(directory, 'data.json')
  version_data = data_json_load(version_file_path)
  version_data["all_versions"].append(version)
  version_data["current_version"] = version
  data_json_dump(version_file_path, version_data)
  return str(version)

def working_version(directory :str) -> str:
  version_file_path = os.path.join(directory, 'data.json')
  with open(version_file_path, 'r') as version_file:
    file_content = json.load(version_file)
    version = file_content["current_version"]
  return version

def update_working_version(directory :str, new_version :int = 0) -> str:
  if new_version != 0:
    version = new_version
  else:
    version = current_version(directory)
  version_file_path = os.path.join(directory, 'data.json')
  current_version_saved = data_json_load(version_file_path)
  current_version_saved["current_version"] = version
  data_json_dump(version_file_path, current_version_saved)
  return version

def data_json_load(file_path :str) -> dict:
  with open(file_path, 'r') as file:
    data = json.load(file)
  return data

def data_json_dump(file_path :str, data) -> None:
  with open(file_path, 'w') as file:
    json.dump(data, file, indent=2)

def hashing(file_path :str) -> str:
  hash_data = hashlib.sha256()
  with open(file_path, 'rb') as file:
    content = file.read()
  hash_data.update(content)
  return hash_data.hexdigest()

def delete_files(file_path :str) -> None:
  if os.path.exists(file_path):
    os.remove(file_path)
    print(f'Removed files: {file_path}')

def add_files(file_path :str, snapshot_path :str ,file_hash :str) -> None:
  file_version = os.path.join(snapshot_path, file_hash)
  with gzip.open(file_version, 'rb') as file_content:
    content = pickle.load(file_content)
  directory = os.path.dirname(file_path)
  if not os.path.exists(directory):
    os.makedirs(directory, exist_ok=True)
  with open(file_path, 'wb') as f:
    f.write(content)


from setuptools import setup

with open("docs/README_PYPI.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="snapvc",
    version="1.2.1",
    author="Shreyash Mogaveera",
    author_email="shreyashmogaveera@gmail.com",
    description="A lightweight version control system with cross-platform support",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ShreyashM17/SnapVC",
    project_urls={
        "Bug Tracker": "https://github.com/ShreyashM17/SnapVC/issues",
        "Documentation": "https://github.com/ShreyashM17/SnapVC#readme",
        "Source Code": "https://github.com/ShreyashM17/SnapVC",
    },
    packages=["snapvc"],
    entry_points={
        "console_scripts": [
            "svcs=snapvc.main:main",
        ],
    },
    classifiers=[
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Topic :: Software Development :: Version Control",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Operating System :: OS Independent",
        "Environment :: Console",
    ],
    keywords=[
        "version-control",
        "vcs", 
        "educational",
        "git",
        "snapshots",
        "cross-platform",
        "cli",
        "learning",
        "development-tools"
    ],
    python_requires=">=3.7",
    license="MIT",
    zip_safe=False,
)
